@@ -171,6 +171,7 @@ soup_message_cleanup (SoupMessage *req)
 	g_return_if_fail (req != NULL);
 
 	if (req->connection && 
+	    soup_connection_is_keep_alive (req->connection) &&
 	    req->priv->read_tag &&
 	    req->status == SOUP_STATUS_READING_RESPONSE) {
 		soup_transfer_read_set_callbacks (req->priv->read_tag,
@@ -292,6 +293,8 @@ void
 soup_message_cancel (SoupMessage *msg) 
 {
 	soup_message_set_error (msg, SOUP_ERROR_CANCELLED);
+	if (msg->connection)
+		soup_connection_set_keep_alive (msg->connection, FALSE);
 	soup_message_issue_callback (msg);
 }
 
@@ -767,7 +770,7 @@ static void
 authorize_handler (SoupMessage *msg, gboolean proxy)
 {
 	const GSList *vals;
-	SoupAuth *auth, *old_auth;
+	SoupAuth *auth;
 	SoupContext *ctx;
 	const SoupUri *uri;
 
@@ -780,18 +783,56 @@ authorize_handler (SoupMessage *msg, gboolean proxy)
 					             "WWW-Authenticate");
 	if (!vals) goto THROW_CANT_AUTHENTICATE;
 
-        auth = soup_auth_new_from_header_list (uri, vals);
+	auth = soup_auth_lookup (ctx);
+	if (auth) {
+		g_assert (auth->status != SOUP_AUTH_STATUS_INVALID);
+
+		if (auth->status == SOUP_AUTH_STATUS_PENDING) {
+			if (auth->controlling_msg == msg) {
+				auth->status = SOUP_AUTH_STATUS_FAILED;
+				goto THROW_CANT_AUTHENTICATE;
+			}
+			else {
+				soup_message_requeue (msg);
+				return;
+			}
+		}
+		else if (auth->status == SOUP_AUTH_STATUS_FAILED ||
+			 auth->status == SOUP_AUTH_STATUS_SUCCESSFUL) {
+			/*
+			 * We've failed previously, but we'll give it
+			 * another go, or we've been successful
+			 * previously, but it's not working anymore.
+			 *
+			 * Invalidate the auth, so it's removed from the
+			 * hash and try it again as if we never had it
+			 * in the first place.
+			 */
+			soup_auth_invalidate (auth, ctx);
+			soup_message_requeue (msg);
+			return;
+		}
+	}
+
 	if (!auth) {
-		soup_message_set_error_full (
-			msg, 
-			proxy ? 
-			        SOUP_ERROR_CANT_AUTHENTICATE_PROXY : 
-			        SOUP_ERROR_CANT_AUTHENTICATE,
-			proxy ? 
-			        "Unknown authentication scheme required by "
-			        "proxy" :
-			        "Unknown authentication scheme required");
-		return;
+		auth = soup_auth_new_from_header_list (uri, vals);
+
+		if (!auth) {
+			soup_message_set_error_full (
+				msg, 
+				proxy ? 
+			                SOUP_ERROR_CANT_AUTHENTICATE_PROXY : 
+			                SOUP_ERROR_CANT_AUTHENTICATE,
+				proxy ? 
+			                "Unknown authentication scheme "
+				        "required by proxy" :
+			                "Unknown authentication scheme "
+				        "required");
+			return;
+		}
+
+		auth->status = SOUP_AUTH_STATUS_PENDING;
+		auth->controlling_msg = msg;
 	}
 
 	/*
@@ -809,24 +850,15 @@ authorize_handler (SoupMessage *msg, gboolean proxy)
 	}
 
 	/*
-	 * Initialize with auth data (possibly returned from auth callback).
+	 * Initialize with auth data (possibly returned from 
+	 * auth callback).
 	 */
 	soup_auth_initialize (auth, uri);
 
-	if (auth->type == SOUP_AUTH_TYPE_NTLM)
-		old_auth = msg->connection->auth;
-	else
-		old_auth = soup_auth_lookup (ctx);
-
-	if (old_auth) {
-		if (!soup_auth_invalidates_prior (auth, old_auth)) {
-			soup_auth_free (auth);
-			goto THROW_CANT_AUTHENTICATE;
-		}
-	}
-
 	if (auth->type == SOUP_AUTH_TYPE_NTLM) {
-		if (old_auth) 
+		SoupAuth *old_auth = soup_auth_lookup (ctx);
+
+		if (old_auth && auth != old_auth) 
 			soup_auth_free (old_auth);
 		msg->connection->auth = auth;
 	} else
@@ -843,6 +875,29 @@ authorize_handler (SoupMessage *msg, gboolean proxy)
 			                SOUP_ERROR_CANT_AUTHENTICATE);
 }
 
+static void
+validate_authorize_handler (SoupMessage *msg, gpointer user_data)
+{
+	SoupContext *ctx;
+	SoupAuth *auth;
+
+	ctx = soup_get_proxy ();
+	if (ctx) {
+		auth = soup_auth_lookup (ctx);
+		if (auth) {
+			auth->status = SOUP_AUTH_STATUS_SUCCESSFUL;
+			auth->controlling_msg = NULL;
+		}
+	}
+
+	ctx = msg->context;
+	auth = soup_auth_lookup (ctx);
+	if (auth) {
+		auth->status = SOUP_AUTH_STATUS_SUCCESSFUL;
+		auth->controlling_msg = NULL;
+	}
+}
+
 static void 
 redirect_handler (SoupMessage *msg, gpointer user_data)
 {
@@ -944,6 +999,16 @@ static SoupHandlerData global_handlers [] = {
 		RESPONSE_ERROR_CODE_HANDLER, 
 		{ 407 }
 	},
+	/*
+	 * Handle response codes 2xx for validating auths.
+	 */
+	{
+		SOUP_HANDLER_PRE_BODY,
+		(SoupCallbackFn) validate_authorize_handler,
+		NULL,
+		RESPONSE_ERROR_CLASS_HANDLER,
+		{ SOUP_ERROR_CLASS_SUCCESS }
+	},
 	{ 0 }
 };
 
